import torch
from transformers import AutoProcessor, AutoModelForZeroShotImageClassification
from PIL import Image, UnidentifiedImageError
import numpy as np
import requests
from io import BytesIO
import os
from pathlib import Path
from hashlib import md5
import tqdm
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
import clip

class ImageArrayDataset(Dataset):
    def __init__(self, list_of_images):
        self.images = list_of_images

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        images = self.images[idx]
        return images

class CaptioningDataset(Dataset):
    def __init__(self, captions):
        self.caption = captions

    def __len__(self):
        return len(self.caption)

    def __getitem__(self, idx):
        caption = self.caption[idx]
        return caption



class PLIP_ViT:
    def __init__(self, model_name, device=None, cache_dir=None):
        self.model_name = model_name
        self.device = device or ("cuda" if torch.cuda.is_available() else "cpu")

        processor = AutoProcessor.from_pretrained("vinid/plip", cache_dir=cache_dir)
        def preprocess_wrapper(img):
            result = processor.image_processor(img)['pixel_values']
            return result
        self.preprocess = preprocess_wrapper

        self.model = self._load_model(cache_dir)
        

    def _load_model(self, cache_dir):
        if self.model_name == "PLIP-ViT-B-32":
            model = AutoModelForZeroShotImageClassification.from_pretrained("vinid/plip", cache_dir=cache_dir)
            return model.to(self.device)
        else:
            raise ValueError(f"Model {self.model_name} not supported.")
            

    def preprocess_single_image(self, img):
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.124 Safari/537.36'
        }
        try:
            # Convert local or http URL to Image
            if isinstance(img, str):
                if os.path.exists(img):  # Local URL
                    img = Image.open(img)
                elif img.startswith('http'):  # Remote URL
                    response = requests.get(img, headers=headers)
                    img = Image.open(BytesIO(response.content))
                else:
                    raise ValueError("Invalid URL or path")

            # Convert numpy array to PIL Image
            elif isinstance(img, np.ndarray) and img.ndim == 3:
                img = Image.fromarray(img)

            ## Check if the image is corrupted
            #img.verify()

            # Now, preprocess the PIL Image
            result = self.preprocess(img)
            return result

        except (requests.RequestException, ValueError, UnidentifiedImageError, FileNotFoundError) as e:
            raise e

    def preprocess_images(self, imgs):
        corrupted_indices = []

        # Handle batched numpy arrays (N * W * H * 3)
        if isinstance(imgs, np.ndarray) and imgs.ndim == 4:
            n_imgs = len(imgs)
            processed_imgs = []
            for idx, img in enumerate(imgs):
                try:
                    processed_imgs.append(self.preprocess_single_image(img))
                except Exception:
                    corrupted_indices.append(idx)

        # For other types of input (list or single input)
        else:
            if not isinstance(imgs, list):
                imgs = [imgs]
            n_imgs = len(imgs)

            processed_imgs = []
            for idx, img in enumerate(imgs):
                try:
                    processed_imgs.append(self.preprocess_single_image(img))
                except Exception:
                    corrupted_indices.append(idx)

        if corrupted_indices:
            return n_imgs, corrupted_indices

        return n_imgs, processed_imgs

    

    def embed_images(self, imgs, num_workers=1, batch_size=32, normalize=True):
        number_of_images, processed_imgs_or_errors = self.preprocess_images(imgs)
        
        if isinstance(processed_imgs_or_errors[0], int):  # It's an error list
            if number_of_images == 1:
                raise ValueError("The image is inaccessible or corrupted")
            else:
                raise ValueError(f"Some images were inaccessible or corrupted. Indices: {', '.join(map(str, processed_imgs_or_errors))}")
            return
        else:
            list_of_preprocessed_arrays = processed_imgs_or_errors

        dataset = ImageArrayDataset(list_of_preprocessed_arrays)
        dataloader = DataLoader(dataset, batch_size=batch_size, num_workers=num_workers)
        image_embeddings = []
        total = len(list_of_preprocessed_arrays) // batch_size
        pbar = tqdm.tqdm(total=total, position=0)
        with torch.no_grad():
            for images in dataloader:
                if isinstance(images, list):
                    images = torch.cat(images)
                images = images.to(self.device)
                embedding = self.model.get_image_features(images).detach().cpu().numpy()
                image_embeddings.extend(embedding)
                pbar.update(1)
            pbar.close()
        image_embeddings = np.array(image_embeddings)
        # Assuming normalize_embedding is a member variable you want to add, else remove this
        if normalize:
            image_embeddings = image_embeddings / np.linalg.norm(image_embeddings, axis=1, keepdims=True)
        return image_embeddings
    

    def embed_texts(self, list_of_labels, num_workers=1, batch_size=32, normalize=True):
        # If the input is a single string, convert it into a list
        if isinstance(list_of_labels, str):
            list_of_labels = [list_of_labels]
            
        dataset = CaptioningDataset(list_of_labels)
        dataloader = DataLoader(dataset, batch_size=batch_size, num_workers=num_workers)
        text_embeddings = []
        total = len(list_of_labels) // batch_size
        pbar = tqdm.tqdm(total=total, position=0)
        with torch.no_grad():
            for captions in dataloader:
                tkn = clip.tokenize(captions, truncate=True).to(self.device)
                embedding = self.model.get_text_features(tkn).detach().cpu().numpy()
                text_embeddings.extend(embedding)
                pbar.update(1)
            pbar.close()
        text_embeddings = np.array(text_embeddings)
        if normalize:
            text_embeddings = text_embeddings / np.linalg.norm(text_embeddings, axis=1, keepdims=True)
        return text_embeddings



import numpy as np
from PIL import Image
from tqdm import tqdm
import pandas as pd
from sklearn.metrics import *
from PIL import Image
import random
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import DataLoader
import os
opj = os.path.join


class AverageMeter(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count



class ImageDataset_with_label(torch.utils.data.Dataset):
    def __init__(self,
                 image_list,
                 label_list,
                 preprocess=None):
        self.image_list = image_list
        self.label_list = label_list
        self.preprocess = preprocess

    def __len__(self):
        return len(self.image_list)

    def __getitem__(self, idx):
        # Load the image
        image_path = self.image_list[idx]
        label = self.label_list[idx]
        image = Image.open(image_path).convert("RGB")
        if self.preprocess is not None:
            image = self.preprocess(image)
        return image, label
    

class BackboneProber(nn.Module):
    def __init__(self, backbone, classifier):
        super(BackboneProber, self).__init__()
        self.backbone = backbone
        self.classifier = classifier

    def forward(self, x):
        x = x.to(self.backbone.vision_model.encoder.layers[0].mlp.fc1.weight.dtype)
        embedding = self.backbone.get_image_features(x) # get embedding via Huggingface approach
        out = self.classifier(embedding)
        out = out.squeeze()
        return out, embedding
    

class FineTuner():
    def __init__(self,
                 backbone,
                 preprocess,
                 num_classes,
                 freeze_vit=False,
                 random_state=42,
                 checkpoint=None,
                 device=None
                 ):
        super(FineTuner, self).__init__()
        backbone = backbone
        self.preprocess = preprocess
        self.num_classes = num_classes
        self.freeze_vit = freeze_vit
        self.random_state = random_state
        self.device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        self.loss_func = torch.nn.CrossEntropyLoss()  # this is for regression mean squared loss

        self.in_features = backbone.visual_projection.out_features
        classifier = nn.Linear(in_features = self.in_features,
                                    out_features = self.num_classes)
        self.model = BackboneProber(backbone, classifier)
        if checkpoint is not None:
            self.load_model(checkpoint)
        self.train_init()
    
    def predict(self, image):
        # image can be either PIL Image or a list of PIL Image
        x = self.preprocess(image)
        # Convert the list of NumPy arrays to a PyTorch tensor
        tensor_list = [torch.tensor(arr) for arr in x]
        # Convert the list of PyTorch tensors to a single tensor
        x = torch.stack(tensor_list)
        out, _ = self.model(x)
        return out
    
    def extract_embedding(self, image):
        # image can be either PIL Image or a list of PIL Image
        x = self.preprocess(image)
        # Convert the list of NumPy arrays to a PyTorch tensor
        tensor_list = [torch.tensor(arr) for arr in x]
        # Convert the list of PyTorch tensors to a single tensor
        x = torch.stack(tensor_list)
        _, embedding = self.model(x)
        return embedding
        
    def train_init(self):
        np.random.seed(self.random_state)
        random.seed(self.random_state)
        torch.manual_seed(self.random_state)
        torch.cuda.manual_seed(self.random_state)
        torch.cuda.manual_seed_all(self.random_state)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    def train(self,
              df,
              validation_split=0.1,
              batch_size=32,
              num_workers=1,
              lr=1e-5,
              nepochs=100
              ):
        train_loader, test_loader = self.prepare_dataloader(df, validation_split, batch_size, num_workers)
        self.train_nn(train_loader,
                      test_loader,
                      lr=lr,
                      nepochs=nepochs
                      )

    def prepare_dataloader(self,
                           df,
                           validation_split,
                           batch_size,
                           num_workers):

        image_list = np.array(df['image'].to_list())
        label_list  = np.array(df['label'].to_list())

        
        np.random.seed(self.random_state)
        order = np.arange(len(image_list))
        np.random.shuffle(order)
        
        image_list_shuffled = image_list[order]
        label_list_shuffled = label_list[order]
        
        print("Preparing dataloader ...")
        train_idx = order[0:int(len(order)*(1-validation_split))]
        image_dataset = ImageDataset_with_label(image_list_shuffled[train_idx], label_list_shuffled[train_idx], preprocess=self.preprocess)
        train_loader = DataLoader(image_dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers)
        
        if validation_split > 0:
            test_idx = order[int(len(order)*(1-validation_split)):]
            image_dataset = ImageDataset_with_label(image_list_shuffled[test_idx], label_list_shuffled[test_idx], preprocess=self.preprocess)
            test_loader = DataLoader(image_dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers)
        else:
            test_loader = None

        return train_loader, test_loader
    

    def train_nn(self,
                 train_loader,
                 test_loader=None,
                 evaluation_step=1,
                 lr=1e-5,
                 nepochs=5,
                 ):
        self.model = self.model.to(self.device)

        if self.freeze_vit:
            optimizer = torch.optim.AdamW(self.model.classifier.parameters(), lr=lr)
        else:
            optimizer = torch.optim.AdamW(self.model.parameters(), lr=lr)
        
        train_df = pd.DataFrame()
        test_df = pd.DataFrame()
        for epoch in range(nepochs):
            loss_all = 0
            y_true = np.array([])
            y_pred = np.array([])
            loss_log = AverageMeter()
            loss = 0
            tqdm_loader = tqdm(train_loader)
            for x, y in tqdm_loader:
                tqdm_loader.set_description(f"Last Batch -- loss: {loss:.2f}")
                x = torch.cat(x)
                x = x.to(self.device)
                y = y.reshape(-1).to(self.device)
                self.model.train()
                prediction, _ = self.model(x)
                loss = self.loss_func(prediction, y)     # must be (1. nn output, 2. target)
                proba = F.softmax(prediction, dim=1)
                y_hat = prediction.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
                y_true = np.concatenate((y_true, y.data.cpu().numpy().reshape(-1)))
                y_pred = np.concatenate((y_pred, y_hat.cpu().numpy().reshape(-1)))
                loss_log.update(loss)
                loss_all += loss.data.cpu().numpy()
                optimizer.zero_grad()   # clear gradients for next train
                loss.backward()         # backpropagation, compute gradients
                optimizer.step()        # apply gradients
            
            acc_train = accuracy_score(y_true, y_pred)
            f1_train = f1_score(y_true, y_pred, average='weighted')
            res_dict_train = {'loss': loss_all, 'accuracy': acc_train, 'f1_weighted': f1_train}
            row_train = pd.DataFrame(res_dict_train, index=[epoch])
            train_df = pd.concat([train_df, row_train], axis=0)
            print(f'epoch: {epoch+1:04d} \t train loss: {loss_all:.4e} \t ACC_train: {acc_train:.2f} \t F1_train: {f1_train:.2f}')

            if (epoch+1) % evaluation_step == 0 and test_loader is not None:
                self.model.eval()
                # plot and show learning process
                y_val_true, y_val_pred, loss_val_all = self.test_nn(test_loader)
                acc_val = accuracy_score(y_val_true, y_val_pred)
                f1_val = f1_score(y_val_true, y_val_pred, average='weighted')
                res_dict_val = {'loss': loss_val_all, 'accuracy': acc_val, 'f1_weighted': f1_val}
                row_test = pd.DataFrame(res_dict_val, index=[epoch])
                test_df = pd.concat([test_df, row_test], axis=0)
                print(f'epoch: {epoch+1:04d} \t validation loss: {loss_val_all:.4e} \t ACC_val: {acc_val:.2f} \t F1_val: {f1_val:.2f}')
        return
        
    def test_nn(self,
                test_loader,
                progress=True,
                calc_loss=True,
                ):
        self.model.eval()
        self.model = self.model.to(self.device)
        
        with torch.no_grad():
            loss_all = 0
            y_true = np.array([])
            y_pred = np.array([])
            if progress:
                pbar = tqdm(total=len(test_loader))
            for x, y in test_loader:
                x = torch.cat(x)
                x = x.to(self.device)
                y = y.reshape(-1).to(self.device)
                if progress:
                    pbar.update()
                
                prediction, _ = self.model(x)
                proba = F.softmax(prediction, dim=1)
                y_hat = prediction.argmax(dim=1, keepdim=True)  # get the index of the max log-probability
                y_true = np.concatenate((y_true, y.data.cpu().numpy().reshape(-1)))
                y_pred = np.concatenate((y_pred, y_hat.data.cpu().numpy().reshape(-1)))
                if calc_loss:
                    loss = self.loss_func(prediction, y)     # must be (1. nn output, 2. target)
                    loss_all += loss.data.cpu().numpy()
                #print(pd.DataFrame(np.c_[y_true, y_pred], columns=['y_true', 'y_pred']))
        return y_true, y_pred, loss_all
    
    def save_model(self, savedir):
        torch.save(self.model.cpu().state_dict(), savedir)

    def load_model(self, checkpoint):
        ckpt = torch.load(checkpoint)
        ckpt_backbone = OrderedDict((key.replace('backbone.', ''), value) for key, value in ckpt.items() if key.startswith('backbone.'))
        ckpt_classifier = OrderedDict((key.replace('classifier.', ''), value) for key, value in ckpt.items() if key.startswith('classifier.'))
        self.model.backbone.load_state_dict(ckpt_backbone)
        self.model.classifier.load_state_dict(ckpt_classifier)

    